#!/usr/bin/python
##########################################
# program: solve a labyrinth             #
# file: LabyrinthSolving                 #
# Author: Th. Kaffka, Cologne, Germany   #
# Date: 27.11.2020                       #
##########################################
import sys
sys.path.append("/home/pi/Ae/Roboter/Interface")
sys.path.append("/home/pi/Ae/Roboter/lib")
import brickpi3 
import time
from LegoMotor import LegoMotor
from LegoVehicle import *
from LegoTouchSensor import LegoTouchSensor
from HitechnicCompassSensor import HitechnicCompassSensor
from LegoColorSensor import LegoColorSensor
import tool4robot as t4r

BP = brickpi3.BrickPi3()

# define the hardware.
motor_l = LegoMotor(BP.PORT_C)
motor_r = LegoMotor(BP.PORT_B)
ts = LegoTouchSensor(BP.PORT_3)
cls = LegoColorSensor(BP.PORT_4)
cs = HitechnicCompassSensor(BP.PORT_2)
vehicle = LegoVehicle(WHEEL_SIZE_EV3, 11.0, motor_l, motor_r)
time.sleep(10)

#define constants.
COLOR_WALL = 6
COLOR_AREA = 7
TURN_LEFT = -1
TURN_RIGHT = 1

# symbols.
NO_DATA = 0
EXIT_FOUND = 1
NO_EXIT_FOUND = 2
WALL = 3
TIME_OVER = 4
MOVE_UP = 5
MOVE_BACK_LONG = 6
MOVE_BACK_SHORT = 7
MOVE_LEFT = 8
MOVE_RIGHT = 9
BOT_STOP = 10
BOT_UP = 11
BOT_BACK_LONG = 12
BOT_BACK_SHORT = 13
BOT_LEFT = 14
BOT_RIGHT = 15

# parameter.
STACK_LEN = 10 # the length of the stack.
SPEED_LAB = 27 # motor speed for labyrinth.
SPEED_ROT = 15 # motor speed for rotation.
BACK_LONG = 7 # long back distance.
BACK_SHORT = 4 # short back distance.
TIMEOUT_BORDER  = 800 # border for time out.
COMPASS_CORR = 3 # factor for compass direction finding.

# define global vars.
comp_direction = [0, 0, 0, 0]
compass_ndx = 0
goal_stack = []
top_stack = -1
maze_end = False

#############################
# define the main directions.
def initDirection():
    comp_direction[0] = cs.getDirection();
    comp_direction[1] = t4r.corrDirection(comp_direction[0] + 90);
    comp_direction[2] = t4r.corrDirection(comp_direction[0] + 180);
    comp_direction[3] = t4r.corrDirection(comp_direction[0] + 270);

###################
# set up the robot.
def setUp():
    global goal_stack

    # prepare the vehicle.
    initDirection()
    vehicle.resetTachoCount()
    vehicle.setSpeed(SPEED_LAB)
    
    for i in range (0, STACK_LEN):
        goal_stack += [0]

    pushStack(MOVE_UP)

##############################
# push an item onto the stack.
def pushStack(item):
    global top_stack
    top_stack += 1
    if (top_stack == STACK_LEN):
        top_stack = STACK_LEN-1
    goal_stack[top_stack] = item

################################
# pop an item from top of stack.
def popStack():
    global top_stack
    if (top_stack == -1):
        return -1
    item = goal_stack[top_stack]
    top_stack -= 1
    if (top_stack < 0):
        top_stack = -1
    return item

########################################
# tests, if the linesensor is on a wall.
def getLightStatus():
    light = cls.getColor()
    if (light == COLOR_WALL):
        return COLOR_WALL
    return COLOR_AREA

###########################################
# the artificial intelligence of the robot.
def ruleBasis():
    global top_stack, maze_end
    stack_value = goal_stack[top_stack]
    popStack()
    if (stack_value == EXIT_FOUND):
        move_engine(BOT_STOP)
        maze_end = True
    elif (stack_value == NO_EXIT_FOUND):
        move_engine(BOT_STOP)
        maze_end = True
    elif (stack_value == WALL):
        pushStack(MOVE_UP)
        pushStack(MOVE_BACK_SHORT)
        pushStack(MOVE_RIGHT)
        pushStack(MOVE_BACK_LONG)
    elif (stack_value == TIME_OVER):
        pushStack(MOVE_UP)
        pushStack(MOVE_BACK_SHORT)
        pushStack(MOVE_LEFT)
        pushStack(MOVE_BACK_SHORT)
    elif (stack_value == MOVE_UP):
        move_engine(BOT_UP)
    elif (stack_value == MOVE_BACK_LONG):
        move_engine(BOT_BACK_LONG)
    elif (stack_value == MOVE_BACK_SHORT):
        move_engine(BOT_BACK_SHORT)
    elif (stack_value == MOVE_LEFT):
        move_engine(BOT_LEFT)
    elif (stack_value == MOVE_RIGHT):
        move_engine(BOT_RIGHT)
    elif (stack_value == NO_DATA):
        top_stack = -1

##################
# moves the robot.
def move_engine(command):
    if (command == BOT_STOP):
        vehicle.stop()
    elif (command == BOT_UP):
        pushStack(move_since_wall())
    elif (command == BOT_BACK_LONG):
        move_back(BACK_LONG)
    elif (command == BOT_BACK_SHORT):
        move_back(BACK_SHORT);
    elif (command == BOT_LEFT):
        rotate(TURN_LEFT)
    elif (command == BOT_RIGHT):
        rotate(TURN_RIGHT)

#####################################################
# move since a wall is reached, or a time out occurs.
def move_since_wall():
    vehicle.setSpeed(SPEED_LAB)
    vehicle.resetTachoCount()
    vehicle.forward()
    result = NO_DATA
    while (result == NO_DATA):
        vehicle.countTacho()
        cs.getDirection() # dummy.
        if (ts.isTouched()):
            result = EXIT_FOUND
        elif (getLightStatus() == COLOR_WALL):
            result = WALL
        elif (motor_l.getTachoCount() > TIMEOUT_BORDER
              or motor_r.getTachoCount() > TIMEOUT_BORDER):
            result = TIME_OVER
    vehicle.stop()
    return result

#######################
# moves the robot back.
def move_back(len):
    vehicle.setSpeed(SPEED_LAB)
    vehicle.backward(len)
    while vehicle.isMoving():
        cs.getDirection() # dummy.

###############################
# rotate only in one direction.
def rotateOnly(dir):
    if (dir == TURN_LEFT):
        vehicle.rotateLeft()
    elif (dir == TURN_RIGHT):
        vehicle.rotateRight()

#######################################
# rotate to the next compass direction.
def rotate(dir):
    global maze_end, compass_ndx
    vehicle.setSpeed(SPEED_ROT)
    if (dir == TURN_RIGHT):
        compass_ndx += 1
        if (compass_ndx > 3):
            compass_ndx = 0
    elif (dir == TURN_LEFT):
        compass_ndx -= 1
        if (compass_ndx < 0):
            compass_ndx = 3
    direction_low = comp_direction[compass_ndx] - COMPASS_CORR
    direction_high = comp_direction[compass_ndx] + COMPASS_CORR
    compass_dir = cs.getDirection()
    rotateOnly(dir)
    while (not(compass_dir >= direction_low 
          and compass_dir <= direction_high) and not maze_end):
        if (ts.isTouched()):
            maze_end = True
        compass_dir = cs.getDirection()
    vehicle.stop()

print ("Start")

try:
    t4r.waitUntilTouch(ts)
    time.sleep(2)

    setUp()

    while (not maze_end):
        ruleBasis()

except KeyboardInterrupt:
    pass
finally:
    vehicle.stop()
    BP.reset_all()
    
print ("Ende")

